/*	$Id: TimeOfDay.h 2 2005-11-11 21:32:03Z faxguy $ */
/*
 * Copyright (c) 1990-1996 Sam Leffler
 * Copyright (c) 1991-1996 Silicon Graphics, Inc.
 * HylaFAX is a trademark of Silicon Graphics
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#ifndef _TimeOfDay_
#define	_TimeOfDay_
/*
 * Time Of Day Support.
 */
#include "Types.h"
#include <time.h>

// NB: should be a nested class
class _tod {
private:
    int nextDay(int d, int wday) const;
public:
    int		days;		// bit vector of ok days
    time_t	start, end;	// time range (hours*60+minutes)
    _tod*	next;		// linked list

    _tod(int days, time_t start, time_t end);
    _tod(const _tod& other);

    time_t nextTime(int wday, time_t hm) const;
};

class TimeOfDay {
private:
    _tod	tod;

    void add(int days, time_t start, time_t end);
public:
    TimeOfDay();
    TimeOfDay(const TimeOfDay& other);
    virtual ~TimeOfDay();

    void reset();
    void parse(const char*);

    time_t nextTimeOfDay(time_t) const;
};
#endif /* _TimeOfDay_ */
