/*	$Id: ondelay.c 823 2008-04-26 22:34:29Z faxguy $ */
/*
 * Copyright (c) 1993-1996 Sam Leffler
 * Copyright (c) 1993-1996 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>

#include "port.h"

#ifndef O_NDELAY
#ifdef O_NONBLOCK
#define O_NDELAY	O_NONBLOCK
#else
#define	O_NDELAY	0
#endif
#endif

int
main(int ac, char* av[])
{
    close(0);
    if (open(av[1], O_RDWR|O_NDELAY) < 0) {
	perror(av[1]);
	return (-1);
    }
    (void) fcntl(0, F_SETFL, fcntl(0, F_GETFL, 0) &~ O_NDELAY);
#ifndef ONDELAY_NOSTDINDUP
    close(1); if (dup(0) < 0) return (-1);	/* for Ultrix and other old BSD stty(1)s */
#endif
    execvp(av[2], &av[2]);
    perror(av[2]);
    return (-1);
}
