/*	$Id: FaxItem.c++ 31 2005-12-21 01:17:11Z faxguy $ */
/*
 * Copyright (c) 1990-1996 Sam Leffler
 * Copyright (c) 1991-1996 Silicon Graphics, Inc.
 * HylaFAX is a trademark of Silicon Graphics
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include "FaxRequest.h"

FaxItem::FaxItem(FaxSendOp o, u_short d, const fxStr& a, const fxStr& s)
    : item(s)
    , addr(a)
{
    op = o;
    dirnum = d;
}
FaxItem::FaxItem() {}
FaxItem::FaxItem(const FaxItem& other)
    : item(other.item)
    , addr(other.addr)
{
    op = other.op;
    dirnum = other.dirnum;
}
FaxItem::~FaxItem() {}

int FaxItem::compare(FaxItem const*) const { return 0; }

bool
FaxItem::isSavedOp() const
{
    return (op == FaxRequest::send_tiff_saved ||
	    op == FaxRequest::send_postscript_saved ||
	    op == FaxRequest::send_pcl_saved ||
	    op == FaxRequest::send_pdf_saved ||
	    op == FaxRequest::send_page_saved ||
	    op == FaxRequest::send_data_saved);
}
fxIMPLEMENT_ObjArray(FaxItemArray, FaxItem)
